#include "burp.h"
#include "handy.h"
#include "prog.h"
#include "msg.h"
#include "asyncio.h"
#include "counter.h"
#include "find.h"
#include "berrno.h"

void close_fd(int *fd)
{
	if(*fd<0) return;
	//logp("closing %d\n", *fd);
	close(*fd);
	*fd=-1;
}

void close_fp(FILE **fp)
{
	if(!*fp) return;
	fclose(*fp);
	*fp=NULL;
}

void gzclose_fp(gzFile *fp)
{
	if(!*fp) return;
	gzflush(*fp, Z_FINISH);
	gzclose(*fp);
	*fp=NULL;
}

int is_dir(const char *path)
{
        struct stat buf;

        if(lstat(path, &buf)) return 0;

        return S_ISDIR(buf.st_mode);
}

char *prepend(const char *prep, const char *fname, size_t len, bool slash)
{
	char *rpath=NULL;

	if(prep) len+=strlen(prep)+1;
	if(slash) len++;
	len++;

	if(!(rpath=(char *)malloc(len)))
	{
		logp("could not malloc for relative path of %s\n", rpath);
		return NULL;
	}
	snprintf(rpath, len, "%s%s%s",
		prep?:"", (slash && *fname)?"/":"", fname);
	return rpath;
}

char *prepend_s(const char *prep, const char *fname, size_t len)
{
	if(!prep || !*prep)
	{
		char *ret=NULL;
		if(!(ret=strdup(fname)))
			logp("out of memory in prepend_s\n");
		return ret;
	}
	// Try to avoid getting a double slash in the path.
	if(fname && fname[0]=='/')
	{
		fname++;
		len--;
	}
	return prepend(prep, fname, len, TRUE);
}

int mkpath(char **rpath)
{
	char *cp=NULL;
	struct stat buf;
	//printf("mkpath: %s\n", *rpath);
	if((cp=strrchr(*rpath, '/')))
	{
		*cp='\0';
		if(lstat(*rpath, &buf))
		{
			// does not exist - recurse further down, then come
			// back and try to mkdir it.
			if(mkpath(rpath)) return -1;
			if(mkdir(*rpath, 0777))
			{
				logp("could not mkdir %s: %s\n", *rpath, strerror(errno));
				*cp='/';
				return -1;
			}
		}
		else if(S_ISDIR(buf.st_mode))
		{
			// Is a directory - can put the slash back and return.
		}
		else if(S_ISLNK(buf.st_mode))
		{
			// to help with the 'current' symlink
		}
		else
		{
			// something funny going on
			logp("warning: wanted '%s' to be a directory\n",
			*rpath);
		}
		*cp='/';
	}
	return 0;
}

int build_path(const char *datadir, const char *fname, size_t flen, char **rpath)
{
	if(!(*rpath=prepend_s(datadir, fname, flen))) return -1;
	if(mkpath(rpath))
	{
		if(*rpath) free(*rpath);
		return -1;
	}
	return 0;
}

// return -1 for error, 0 for OK, 1 if the client wants to interrupt the
// transfer.
int do_quick_read(const char *datapth)
{
	int r=0;
	char cmd;
	size_t len=0;
	char *buf=NULL;
	if(async_read_quick(&cmd, &buf, &len)) return -1;

	if(buf)
	{
		if(cmd=='w')
		{
			logp("WARNING: %s\n", buf);
			do_filecounter(cmd, 0);
		}
		else if(cmd=='i')
		{
			// Client wants to interrupt - double check that
			// it is still talking about the file that we are
			// sending.
			if(datapth && !strcmp(buf, datapth))
				r=1;
		}
		else
		{
			logp("unexpected cmd in quick read: %c:%s\n", cmd, buf);
			r=-1;
		}
		free(buf);
	}
	return r;
}

char *get_checksum_str(unsigned char *checksum)
{
	//int i=0;
	//char tmp[3]="";
	static char str[64]="";
/*
	str[0]='\0';
	// Windows does not seem to like me writing it all at the same time.
	// Fuck knows why.
	for(i=0; i<MD5_DIGEST_LENGTH; i++)
	{
		snprintf(tmp, sizeof(tmp), "%02x", checksum[i]);
		strcat(str, tmp);
	}
*/
	snprintf(str, sizeof(str),
	  "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
		checksum[0], checksum[1],
		checksum[2], checksum[3],
		checksum[4], checksum[5],
		checksum[6], checksum[7],
		checksum[8], checksum[9],
		checksum[10], checksum[11],
		checksum[12], checksum[13],
		checksum[14], checksum[15]);
	return str;
}

int write_endfile(unsigned long long bytes, unsigned char *checksum)
{
	int ret=0;
	char endmsg[128]="";

	snprintf(endmsg, sizeof(endmsg), "%llu:%s",
		bytes, get_checksum_str(checksum));

	ret=async_write_str('x', endmsg);
	return ret;
}

#define ZCHUNK 16000

static int do_encryption(EVP_CIPHER_CTX *ctx, unsigned char *inbuf, size_t inlen, unsigned char *outbuf, size_t *outlen, MD5_CTX *md5)
{
	if(!inlen) return 0;
	if(!EVP_CipherUpdate(ctx, outbuf, (int *)outlen, inbuf, (int)inlen))
	{
		logp("Encryption failure.\n");
		return -1;
	}
	if(*outlen>0)
	{
		int ret;
		if(!(ret=async_write('a', (const char *)outbuf, *outlen)))
		{
			if(!MD5_Update(md5, outbuf, *outlen))
			{
				logp("MD5_Update() failed\n");
				return -1;
			}
		}
		return ret;
	}
	return 0;
}

EVP_CIPHER_CTX *enc_setup(int encrypt, const char *encryption_password)
{
	EVP_CIPHER_CTX *ctx=NULL;
	const char *enc_iv="[lkd.$G"; // never change this.

	if(!(ctx=(EVP_CIPHER_CTX *)malloc(sizeof(EVP_CIPHER_CTX))))
	{
		logp("out of memory\n");
		return NULL;
	}
        memset(ctx, 0, sizeof(EVP_CIPHER_CTX));
	// Don't set key or IV because we will modify the parameters.
	EVP_CIPHER_CTX_init(ctx);
	if(!(EVP_CipherInit_ex(ctx, EVP_bf_cbc(), NULL, NULL, NULL, encrypt)))
	{
		logp("EVP_CipherInit_ex failed\n");
		free(ctx);
		return NULL;
	}
	EVP_CIPHER_CTX_set_key_length(ctx, strlen(encryption_password));
	// We finished modifying parameters so now we can set key and IV

	if(!EVP_CipherInit_ex(ctx, NULL, NULL,
		(unsigned char *)encryption_password,
		(unsigned char *)enc_iv, encrypt))
	{
		logp("Second EVP_CipherInit_ex failed\n");
		free(ctx);
		return NULL;
	}
	return ctx;
}

int send_whole_file_gz(const char *fname, const char *datapth, int quick_read, unsigned long long *bytes, const char *encpassword)
{
	int ret=0;
	int zret=0;
	MD5_CTX md5;
#ifdef HAVE_WIN32
	BFILE bfd;
#else
	FILE *fp=NULL;
#endif

	unsigned have;
	z_stream strm;
	int flush=Z_NO_FLUSH;
	unsigned char in[ZCHUNK];
	unsigned char out[ZCHUNK];

	size_t eoutlen;
	unsigned char eoutbuf[ZCHUNK+EVP_MAX_BLOCK_LENGTH];

	EVP_CIPHER_CTX *enc_ctx=NULL;

	if(encpassword && !(enc_ctx=enc_setup(1, encpassword)))
		return -1;

	if(!MD5_Init(&md5))
	{
		logp("MD5_Init() failed\n");
		return -1;
	}


//logp("send_whole_file_gz: %s\n", fname);

#ifdef HAVE_WIN32
	binit(&bfd);
	if(bopen(&bfd, fname, O_RDONLY | O_BINARY | O_NOATIME, 0)<0)
	{
		berrno be;
		logp("Could not open %s: %s\n", fname, be.bstrerror(errno));
		return -1;
	}
#else
	if(!(fp=fopen(fname, "rb")))
	{
		logp("Could not open %s: %s\n", fname, strerror(errno));
		return -1;
	}
#endif

	/* allocate deflate state */
	strm.zalloc = Z_NULL;
	strm.zfree = Z_NULL;
	strm.opaque = Z_NULL;
	if((zret=deflateInit2(&strm, Z_BEST_COMPRESSION, Z_DEFLATED, (15+16),
		8, Z_DEFAULT_STRATEGY))!=Z_OK)

	{
#ifdef HAVE_WIN32
		bclose(&bfd);
#else
		fclose(fp);
#endif
		return -1;
	}

	do
	{
#ifdef HAVE_WIN32
		strm.avail_in=(uint32_t)bread(&bfd, in, ZCHUNK);
#else
		strm.avail_in=fread(in, 1, ZCHUNK, fp);
#endif
		if(strm.avail_in<0)
		{
			logp("Error in read: %d\n", strm.avail_in);
			ret=-1;
			break;
		}
		*bytes+=strm.avail_in;

		// The checksum needs to be later if encryption is being used.
		if(!enc_ctx)
		{
			if(!MD5_Update(&md5, in, strm.avail_in))
			{
				logp("MD5_Update() failed\n");
				ret=-1;
				break;
			}
		}

		if(strm.avail_in) flush=Z_NO_FLUSH;
		else flush=Z_FINISH;

		strm.next_in=in;

		/* run deflate() on input until output buffer not full, finish
			compression if all of source has been read in */
		do
		{
			strm.avail_out = ZCHUNK;
			strm.next_out = out;
			zret = deflate(&strm, flush); /* no bad return value */
			if(zret==Z_STREAM_ERROR) /* state not clobbered */
			{
				logp("z_stream_error\n");
				ret=-1;
				break;
			}
			have = ZCHUNK-strm.avail_out;

			if(enc_ctx)
			{
				if(do_encryption(enc_ctx, out, have, eoutbuf, &eoutlen, &md5))
				{
					ret=-1;
					break;
				}
			}
			else if(async_write('a', (const char *)out, have))
			{
				ret=-1;
				break;
			}
			if(quick_read && datapth)
			{
				int qr;
				if((qr=do_quick_read(datapth))<0)
				{
					ret=-1;
					break;
				}
				if(qr) // client wants to interrupt
				{
					flush=Z_FINISH;
					break;
				}
			}
		} while (!strm.avail_out);

		if(ret) break;

		if(strm.avail_in) /* all input will be used */
		{
			ret=-1;
			logp("strm.avail_in=%d\n", strm.avail_in);
			break;
		}
	} while(flush!=Z_FINISH);

	if(!ret)
	{
		if(zret!=Z_STREAM_END)
		{
			logp("ret OK, but zstream not finished: %d\n", zret);
			ret=-1;
		}
		else if(enc_ctx)
		{
			if(!EVP_CipherFinal_ex(enc_ctx,
				eoutbuf, (int *)&eoutlen))
			{
				logp("Encryption failure at the end\n");
				ret=-1;
			}
			else if(eoutlen>0)
			{
			  if(async_write('a', (const char *)eoutbuf, eoutlen))
				ret=-1;
			  else if(!MD5_Update(&md5, eoutbuf, eoutlen))
			  {
				logp("MD5_Update() failed\n");
				ret=-1;
			  }
			}
		}
	}
	deflateEnd(&strm);

#ifdef HAVE_WIN32
	bclose(&bfd);
#else
	fclose(fp);
#endif

	if(enc_ctx)
	{
		EVP_CIPHER_CTX_cleanup(enc_ctx);
		free(enc_ctx);
	}

	if(!ret)
	{
		unsigned char checksum[MD5_DIGEST_LENGTH+1];
		if(!MD5_Final(checksum, &md5))
		{
			logp("MD5_Final() failed\n");
			return -1;
		}

		return write_endfile(*bytes, checksum);
	}
	return ret;
}

int send_whole_file(const char *fname, const char *datapth, int quick_read, unsigned long long *bytes)
{
	int ret=0;
	ssize_t s=0;
	char buf[4096]="";
	MD5_CTX md5;
	if(!MD5_Init(&md5))
	{
		logp("MD5_Init() failed\n");
		return -1;
	}
#ifdef HAVE_WIN32
	BFILE bfd;
	binit(&bfd);
	if(bopen(&bfd, fname, O_RDONLY | O_BINARY | O_NOATIME, 0)<0)
	{
		berrno be;
		logp("Could not open %s: %s\n", fname, be.bstrerror(errno));
		return -1;
	}
	while((s=(uint32_t)bread(&bfd, buf, 4096))>0)
	{
		*bytes+=s;
		if(!MD5_Update(&md5, buf, s))
		{
			logp("MD5_Update() failed\n");
			ret=-1;
			break;
		}
		if(async_write('a', buf, s))
		{
			ret=-1;
			break;
		}
		if(quick_read)
		{
			int qr;
			if((qr=do_quick_read(datapth))<0)
			{
				ret=-1;
				break;
			}
			if(qr)
			{
				// client wants to interrupt
				break;
			}
		}
	}
	bclose(&bfd);
#else
	FILE *fp=NULL;
//printf("send_whole_file: %s\n", fname);
	if(!(fp=fopen(fname, "rb")))
	{
		logp("Could not open %s: %s\n", fname, strerror(errno));
		return -1;
	}
	while((s=fread(buf, 1, 4096, fp))>0)
	{
		*bytes+=s;
		if(!MD5_Update(&md5, buf, s))
		{
			logp("MD5_Update() failed\n");
			ret=-1;
			break;
		}
		if(async_write('a', buf, s))
		{
			ret=-1;
			break;
		}
		if(quick_read)
		{
			int qr;
			if((qr=do_quick_read(datapth))<0)
			{
				ret=-1;
				break;
			}
			if(qr)
			{
				// client wants to interrupt
				break;
			}
		}
	}
	fclose(fp);
#endif
	if(!ret)
	{
		unsigned char checksum[MD5_DIGEST_LENGTH+1];
		if(!MD5_Final(checksum, &md5))
		{
			logp("MD5_Final() failed\n");
			return -1;
		}
		return write_endfile(*bytes, checksum);
	}
	return ret;
}

int set_non_blocking(int fd)
{
    int flags;
    if((flags = fcntl(fd, F_GETFL, 0))<0) flags = 0;
    return fcntl(fd, F_SETFL, flags | O_NONBLOCK);
}
     
int set_blocking(int fd)
{
    int flags;
    if((flags = fcntl(fd, F_GETFL, 0))<0) flags = 0;
    return fcntl(fd, F_SETFL, flags | ~O_NONBLOCK);
}

int do_rename(const char *oldpath, const char *newpath)
{
	if(rename(oldpath, newpath))
	{
		logp("could not rename '%s' to '%s': %s\n",
			oldpath, newpath, strerror(errno)); 
		return -1; 
	}
	return 0;
}

char *get_tmp_filename(const char *basis)
{
	return prepend(basis, ".tmp", strlen(".tmp"), 0 /* no slash */);
}
